function PA7

clear all
close all

% Der Code zur Erstellung der Triangulierung, sowie das finden der Dirichlet-Indizes.
addpath('distmesh') % hier wird der Pfad angegeben, in welchem das 'distmesh'-Paket liegt
% Polyeder fr das Gebiet
pfix = [[-4,-4];[-4,4];[0,4];[0,0];[4,0];[4,-4];[-4,-4]];
% Triangulierung erstellen
[P,T] = distmesh2d(@dpoly,@huniform,0.3,[-5,-5;5,5],pfix,pfix);
% Finde die Dirichlet Knoten
Dirichlet = find(abs(dpoly(P,pfix)) < 1e-3);
% f ist 1 im Teilgebiet D und 0 sonst: verwende f spter um zu bestimmen ob ein Dreieck im Teilgebiet ist oder nicht
f = @(p) (p(:,1) >= -3).*(p(:,1) <= -1).*(p(:,2) >= 1).*(p(:,2) <= 3);

% Initialisierung
nt = 1000; % number of timesteps
tau = 1/10; % timestepping
n = length(P); % Anzahl Diskretisierungspunkte
B = spalloc(n,n,5*n); % Matrix B als sparse Matrix
G = spalloc(n,n,5*n); % Matrix G als sparse Matrix
h = zeros(n,1); % rechten Seite initialisieren

for i = 1:length(T)
    % Wir bestimmen zuerst die Matrix Ainv, womit dann das LGS zur
    % Bestimmung der stk. linearen Funktionen ber dem aktuellen Dreieck
    % fr verschiedene rechte Seiten gelst werden kann (bzw. betrachten
    % wir nur Hutfunktionen, sodass die rechte Seite ein Einheitsvektor ist
    % und die Lsung also die entsprechende Spalte von Ainv ist).
    
    % Determinante von A, vgl. Theorie.
    det = P(T(i,2),1)*P(T(i,3),2) + P(T(i,1),1)*P(T(i,2),2) + P(T(i,1),2)*P(T(i,3),1) ... 
        - P(T(i,1),2)*P(T(i,2),1) - P(T(i,3),1)*P(T(i,2),2) - P(T(i,3),2)*P(T(i,1),1);
    % Damit Ainv angeben
    Ainv = 1/det*[P(T(i,2),1)*P(T(i,3),2) - P(T(i,2),2)*P(T(i,3),1), P(T(i,1),2)*P(T(i,3),1) - P(T(i,1),1)*P(T(i,3),2), P(T(i,1),1)*P(T(i,2),2) - P(T(i,1),2)*P(T(i,2),1);...
                  P(T(i,2),2) - P(T(i,3),2), P(T(i,3),2) - P(T(i,1),2), P(T(i,1),2) - P(T(i,2),2);...
                  P(T(i,3),1) - P(T(i,2),1), P(T(i,1),1) - P(T(i,3),1), P(T(i,2),1) - P(T(i,1),1)];
    % Flcheninahlt des Dreiecks - einerseits relevant fr die Integrale
    % ber Gradient (ist stkw. konstant) - andererseits entspricht 2*vol 
    % der Determinante welche aus der Koordinatentransformation resultiert.
    vol = Inhalt(P(T(i,1),:),P(T(i,2),:),P(T(i,3),:));
              
    % Damit elementweises Aufstellen der Matrix B - p1,p2,p3 sind die 3
    % Hutfunktionen und ihre Koeffizienten stehen in der 1.,2.,3. Spalte von Ainv
    % Hier nur der 1. schon bekannte Teil der BLF!
    B(T(i,1),T(i,1)) = B(T(i,1),T(i,1)) + vol*(Ainv(2,1)^2 + Ainv(3,1)^2); % p1 mit p1
    B(T(i,1),T(i,2)) = B(T(i,1),T(i,2)) + vol*(Ainv(2,1)*Ainv(2,2) + Ainv(3,1)*Ainv(3,2)); % p1 mit p2
    B(T(i,1),T(i,3)) = B(T(i,1),T(i,3)) + vol*(Ainv(2,1)*Ainv(2,3) + Ainv(3,1)*Ainv(3,3)); % p1 mit p3
    B(T(i,2),T(i,1)) = B(T(i,2),T(i,1)) + vol*(Ainv(2,2)*Ainv(2,1) + Ainv(3,2)*Ainv(3,1)); % p2 mit p1 - selbe wie p1 mit p2
    B(T(i,2),T(i,2)) = B(T(i,2),T(i,2)) + vol*(Ainv(2,2)^2 + Ainv(3,2)^2); % p2 mit p2
    B(T(i,2),T(i,3)) = B(T(i,2),T(i,3)) + vol*(Ainv(2,2)*Ainv(2,3) + Ainv(3,2)*Ainv(3,3)); % p2 mit p3
    B(T(i,3),T(i,1)) = B(T(i,3),T(i,1)) + vol*(Ainv(2,3)*Ainv(2,1) + Ainv(3,3)*Ainv(3,1)); % p3 mit p1 - selbe wie p1 mit p3
    B(T(i,3),T(i,2)) = B(T(i,3),T(i,2)) + vol*(Ainv(2,3)*Ainv(2,2) + Ainv(3,3)*Ainv(3,2)); % p3 mit p2 - selbe wie p2 mit p3
    B(T(i,3),T(i,3)) = B(T(i,3),T(i,3)) + vol*(Ainv(2,3)^2 + Ainv(3,3)^2); % p3 mit p3
    
    % Matrix G bauen
    for j = 1:3
        for k = 1:3
            % Baue zuerst das Produkt phi_i*phi_j als Funktion - jede
            % Spalte von Ainv beinhaltet die Polynomkoeffizienten einer Hutfunktion.
            g = @(x,y) Ainv(1,k)*Ainv(1,j) + Ainv(2,k)*Ainv(2,j).*x.^2 + Ainv(3,k)*Ainv(3,j).*y.^2 + ...
                        Ainv(1,k)*Ainv(2,j).*x + Ainv(2,k)*Ainv(1,j).*x + ...
                        Ainv(1,k)*Ainv(3,j).*y + Ainv(3,k)*Ainv(1,j).*y + Ainv(2,k)*Ainv(3,j).*x.*y + Ainv(3,k)*Ainv(2,j).*x.*y;
            % werte nun mit eval_int das Integral aus: transformiere f auf
            % das Referenzdreieck und dort wird eine exakte Quadraturformel
            % verwendet. 2*vol entspricht der Trafodeterminante
            val = 2*vol*eval_int(g,P(T(i,:),:));
            % Anteil zur Matrix G
            G(T(i,j),T(i,k)) = G(T(i,j),T(i,k)) + val;
        end;
    end;
    
    % Rechte Seite h (ist hier zeitunabhngig):
    % zuerst wertet man f in allen Punkten des Dreiecks aus. Wenn alle
    % Punkte innerhalb des Teilgebietes sind (f > 0, bzw. f =1), so ist das Dreieck
    % drin und der Anteil des Dreiecks zur rechten Seite y wird bestimmt
    Punkte = P(T(i,:),:);
    Werte = f(Punkte);
    if isequal(Werte > 0, ones(3,1)) % Das Dreieck ist im Teilgebiet
        % der Flcheninhalt ist der Tetraeder-Inhalt, vgl. Theorieaufgabe
        h(T(i,1)) = h(T(i,1)) + 1/6*abs((P(T(i,2),1)-P(T(i,1),1))*(P(T(i,3),2)-P(T(i,1),2)) - (P(T(i,2),2)-P(T(i,1),2))*(P(T(i,3),1)-P(T(i,1),1)));
        h(T(i,2)) = h(T(i,2)) + 1/6*abs((P(T(i,1),1)-P(T(i,2),1))*(P(T(i,3),2)-P(T(i,2),2)) - (P(T(i,1),2)-P(T(i,2),2))*(P(T(i,3),1)-P(T(i,2),1)));
        h(T(i,3)) = h(T(i,3)) + 1/6*abs((P(T(i,1),1)-P(T(i,3),1))*(P(T(i,2),2)-P(T(i,3),2)) - (P(T(i,1),2)-P(T(i,3),2))*(P(T(i,2),1)-P(T(i,3),1)));
    end;
end;

% Randbehandlung, vgl. Theoriekommentare
G(Dirichlet,:) = 0;
G(:,Dirichlet) = 0;
G = G + sparse(Dirichlet, Dirichlet, ones(size(Dirichlet)), n,n);
% B wird nur auf 0 gesetzt, vgl. Theorie.
B(Dirichlet,:) = 0;
B(:,Dirichlet) = 0;
% RHS mit 1/tau skalieren - unntig hier, da 0-Randwerte
h(Dirichlet) = 0;

% Die 3 verschiedenen Einschrittverfahren
y_expl = zeros(length(B),nt+1);
y_impl = zeros(length(B),nt+1);
y_CN = zeros(length(B),nt+1);
for i = 1:3
    switch i
        case 1 % explizites Euler
            disp('Explizites Euler')
            % Zeitevolution
            for t = 1:nt+1
                y_expl(:,t+1) = y_expl(:,t) + tau*G\(h-B*y_expl(:,t));
            end;
            % plot - da das Verfahren nicht A-stabil ist, sieht man
            % besonders im plot, dass die Approximationen nicht gegen die
            % wahre Lsung konvergieren. Bei den beiden folgenden Verfahren
            % ist dies der Fall (diese sind A-stabil)
            disp('Beginne plot')
            g1 = figure;
            for t = 1:nt+1
                % Es wird nur die Punktwolke geplottet (also die Werte an den Knoten).
                % Es fehlt die Information ber die Dreiecke, ist aber schnell - alternativer plot, wie auf Blatt6
                plot3(P(:,1),P(:,2),y_expl(:,t),'.');
                pause(0.02)
            end;
        case 2 % implizites Euler
            disp('Implizites Euler')
            % Zeitevolution
            for t = 1:nt+1
                y_impl(:,t+1) = (G+tau*B)\(tau*h+G*y_impl(:,t));
            end;
            % plot
            disp('Beginne plot')
            g2 = figure;
            for t = 1:nt+1
                % Es wird nur die Punktwolke geplottet (also die Werte an den Knoten).
                % Es fehlt die Information ber die Dreiecke, ist aber schnell - alternativer plot, wie auf Blatt6
                plot3(P(:,1),P(:,2),y_impl(:,t),'.');
                axis([-4 4 -4 4 0 0.8]);
                pause(0.02)
            end;
        case 3 % Crank-Nicholson
            disp('Crank-Nicholson')
            % Zeitevolution
            for t = 1:nt+1
                y_CN(:,t+1) = (G+tau/2*B)\(tau*h+(G-tau/2*B)*y_CN(:,t));
            end;
            % plot
            disp('Beginne plot')
            g3 = figure;
            for t = 1:nt+1
                % Es wird nur die Punktwolke geplottet (also die Werte an den Knoten).
                % Es fehlt die Information ber die Dreiecke, ist aber schnell - alternativer plot, wie auf Blatt6
                plot3(P(:,1),P(:,2),y_CN(:,t),'.');
                axis([-4 4 -4 4 0 0.8]);
                pause(0.02)
            end;            
    end;
end;

function val = eval_int(f, points)
% - f ist die Funktion welche ber dem Dreieck gegeben durch points integriert werden soll
% - points enthlt die 3 Eckpunkte des Dreiecks
% - Anmerkung: die Trafo-Determinante betrgt 2*vol und wird auerhalb der Funktion dazumultipliziert.

% extrahiere die eckpunkte 
a = points(1,:);
b = points(2,:);
c = points(3,:);
% definiere die transformierte Funktion. Es ist f_transformiert =
% f(Phi(x)), wobei Phi das Referenzdreieck auf das Dreieck gegeben durch points abbildet.
f_transformiert = @(u,v) f(a(1)+u*(b(1)-a(1))+v*(c(1)-a(1)), a(2)+u*(b(2)-a(2))+v*(c(2)-a(2)));
% Damit kann man nun entweder via integral2 über das Refernzdreieck
% integrieren (auskommentiert), oder man verwendet eine in P_2 exakte
% Quadraturformel für das Referenzdreieck
%fy = @(x) 1-x;
%r = integral2(f_transformiert, 0,1, 0,fy);
val = 1/6*(f_transformiert(0.5,0.5) + f_transformiert(0.5,0) + f_transformiert(0,0.5)); % Quadraturformel


function I = Inhalt(p1,p2,p3)
% Flcheninhalt eines Dreiecks ber die Determinante
a = p2-p1;
b = p3-p1;
I = 1/2*abs(a(1)*b(2) - a(2)*b(1));
